@@ -368,7 +368,7 @@ static void dup_header(char *header)
 }
 
 struct digest_parms {
-    const char *realm, *nonce, *opaque;
+    const char *realm, *nonce, *opaque, *domain;
     int rfc2617;
     int send_ainfo;
     int md5_sess;
@@ -389,7 +389,8 @@ struct digest_parms {
         fail_ai_bad_cnonce,
         fail_ai_omit_cnonce,
         fail_ai_omit_digest,
-        fail_ai_omit_nc
+        fail_ai_omit_nc,
+        fail_outside_domain
     } failure;
 };
 
@@ -546,7 +547,8 @@ static int verify_digest_header(struct digest_state *state,
 
     DIGCMP(realm);
     DIGCMP(username);
-    DIGCMP(uri);
+    if (!parms->domain)
+        DIGCMP(uri);
     DIGCMP(nonce);
     DIGCMP(opaque);
     DIGCMP(algorithm);
@@ -558,6 +560,9 @@ static int verify_digest_header(struct digest_state *state,
     if (newstate.cnonce) {
         state->cnonce = ne_strdup(newstate.cnonce);
     }
+    if (parms->domain) {
+        state->uri = ne_strdup(newstate.uri);
+    }
 
     ONN("no digest param given", !state->digest);
 
@@ -647,6 +652,10 @@ static char *make_digest_header(struct digest_state *state,
         ne_buffer_concat(buf, "opaque=\"", parms->opaque, "\", ", NULL);
     }
 
+    if (parms->domain) {
+        ne_buffer_concat(buf, "domain=\"", parms->domain, "\", ", NULL);
+    }
+
     if (parms->failure == fail_req0_stale
         || parms->stale == parms->num_requests) {
         ne_buffer_concat(buf, "stale='true', ", NULL);
@@ -664,7 +673,12 @@ static int serve_digest(ne_socket *sock, void *userdata)
     struct digest_state state;
     char resp[NE_BUFSIZ];
     
-    state.uri = parms->proxy ? "http://www.example.com/fish" : "/fish";
+    if (parms->proxy)
+        state.uri = "http://www.example.com/fish";
+    else if (parms->domain)
+        state.uri = "/fish/0";
+    else
+        state.uri = "/fish";
     state.method = "GET";
     state.realm = parms->realm;
     state.nonce = parms->nonce;
@@ -700,12 +714,30 @@ static int serve_digest(ne_socket *sock, void *userdata)
 
     SEND_STRING(sock, resp);
 
+    /* Give up now if we've sent a challenge which should force the
+     * client to fail immediately: */
+    if (parms->failure == fail_bogus_alg
+        || parms->failure == fail_req0_stale) {
+        return OK;
+    }
+
     do {
+        digest_hdr = NULL;
         CALL(discard_request(sock));
 
-        ONN("no Authorization header sent", digest_hdr == NULL);
-        
-        CALL(verify_digest_header(&state, parms, digest_hdr));
+        if (digest_hdr && parms->domain && (parms->num_requests & 1) != 0) {
+            SEND_STRING(sock, "HTTP/1.1 400 Used Auth Outside Domain\r\n\r\n");
+            return OK;
+        }
+        else if (digest_hdr == NULL && parms->domain && (parms->num_requests & 1) != 0) {
+            /* Do nothing. */
+            NE_DEBUG(NE_DBG_HTTP, "No Authorization header sent, good.\n");
+        }
+        else {
+            ONN("no Authorization header sent", digest_hdr == NULL);
+
+            CALL(verify_digest_header(&state, parms, digest_hdr));
+        }
 
         if (parms->num_requests == parms->stale) {
             state.nonce = ne_concat("stale-", state.nonce, NULL);
@@ -774,27 +806,27 @@ static int digest(void)
 {
     struct digest_parms parms[] = {
         /* RFC 2617-style */
-        { "WallyWorld", "this-is-a-nonce", NULL, 1, 0, 0, 0, 0, 1, 0, fail_not },
-        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", 1, 0, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "this-is-a-nonce", NULL, NULL, 1, 0, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", NULL, 1, 0, 0, 0, 0, 1, 0, fail_not },
         /* ... with A-I */
-        { "WallyWorld", "nonce-nonce-nonce", "opaque-string", 1, 1, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "nonce-nonce-nonce", "opaque-string", NULL, 1, 1, 0, 0, 0, 1, 0, fail_not },
         /* ... with md5-sess. */
-        { "WallyWorld", "nonce-nonce-nonce", "opaque-string", 1, 1, 1, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "nonce-nonce-nonce", "opaque-string", NULL, 1, 1, 1, 0, 0, 1, 0, fail_not },
         /* many requests, with changing nonces; tests for next-nonce handling bug. */
-        { "WallyWorld", "this-is-a-nonce", "opaque-thingy", 1, 1, 0, 0, 1, 20, 0, fail_not },
+        { "WallyWorld", "this-is-a-nonce", "opaque-thingy", NULL, 1, 1, 0, 0, 1, 20, 0, fail_not },
         /* staleness. */
-        { "WallyWorld", "this-is-a-nonce", "opaque-thingy", 1, 1, 0, 0, 0, 3, 2, fail_not },
+        { "WallyWorld", "this-is-a-nonce", "opaque-thingy", NULL, 1, 1, 0, 0, 0, 3, 2, fail_not },
 
 
         /* RFC 2069-style */ 
-        { "WallyWorld", "lah-di-da-di-dah", NULL, 0, 0, 0, 0, 0, 1, 0, fail_not },
-        { "WallyWorld", "fee-fi-fo-fum", "opaque-string", 0, 0, 0, 0, 0, 1, 0, fail_not },
-        { "WallyWorld", "fee-fi-fo-fum", "opaque-string", 0, 1, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "lah-di-da-di-dah", NULL, NULL, 0, 0, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "fee-fi-fo-fum", "opaque-string", NULL, 0, 0, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "fee-fi-fo-fum", "opaque-string", NULL, 0, 1, 0, 0, 0, 1, 0, fail_not },
 
         /* Proxy auth */
-        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", 1, 1, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", NULL, 1, 1, 0, 0, 0, 1, 0, fail_not },
         /* Proxy + A-I */
-        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", 1, 1, 0, 1, 0, 1, 0, fail_not },
+        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", NULL, 1, 1, 0, 1, 0, 1, 0, fail_not },
 
         { NULL }
     };
@@ -828,16 +860,14 @@ static int digest_failures(void)
     };
     size_t n;
 
+    memset(&parms, 0, sizeof parms);
+    
     parms.realm = "WallyWorld";
     parms.nonce = "random-invented-string";
     parms.opaque = NULL;
     parms.rfc2617 = 1;
     parms.send_ainfo = 1;
-    parms.md5_sess = 0;
-    parms.proxy = 0;
-    parms.send_nextnonce = 0;
     parms.num_requests = 1;
-    parms.stale = 0;
 
     for (n = 0; fails[n].message; n++) {
         ne_session *sess = ne_session_create("http", "localhost", 7777);
@@ -899,6 +929,8 @@ static int fail_challenge(void)
           "unknown algorithm in Digest challenge" },
         { "Digest algorithm=MD5-sess, realm=\"foo\"",
           "incompatible algorithm in Digest challenge" },
+        { "Digest algorithm=MD5, qop=auth, nonce=\"foo\", realm=\"foo\", "
+          "domain=\"http://[::1/\"", "could not parse domain" },
 
         /* Multiple challenge failure cases: */
         { "Basic, Digest",
@@ -1009,6 +1041,53 @@ static int multi_handler(void)
     return await_server();
 }
 
+static int domains(void)
+{
+    ne_session *sess;
+    struct digest_parms parms;
+
+    memset(&parms, 0, sizeof parms);
+    parms.realm = "WallyWorld";
+    parms.rfc2617 = 1;
+    parms.nonce = "agoog";
+    parms.domain = "http://localhost:7777/fish/ https://example.com /agaor /other";
+    parms.num_requests = 6;
+
+    CALL(make_session(&sess, serve_digest, &parms));
+
+    ne_set_server_auth(sess, auth_cb, NULL);
+
+    ne_session_proxy(sess, "localhost", 7777);
+
+    CALL(any_2xx_request(sess, "/fish/0"));
+    CALL(any_2xx_request(sess, "/outside"));
+    CALL(any_2xx_request(sess, "/others"));
+    CALL(any_2xx_request(sess, "/fish"));
+    CALL(any_2xx_request(sess, "/fish/2"));
+    CALL(any_2xx_request(sess, "*"));
+    
+    ne_session_destroy(sess);
+
+    return await_server();
+}
+
+static int defaults(void)
+{
+    ne_session *sess;
+    
+    CALL(make_session(&sess, auth_serve, CHAL_WALLY));
+    ne_add_server_auth(sess, NE_AUTH_DEFAULT, auth_cb, NULL);
+    CALL(any_2xx_request(sess, "/norman"));
+    ne_session_destroy(sess);
+    CALL(await_server());
+
+    CALL(make_session(&sess, auth_serve, CHAL_WALLY));
+    ne_add_server_auth(sess, NE_AUTH_ALL, auth_cb, NULL);
+    CALL(any_2xx_request(sess, "/norman"));
+    ne_session_destroy(sess);
+    return await_server();
+}
+
 /* test logout */
 
 /* proxy auth, proxy AND origin */
@@ -1024,5 +1103,7 @@ ne_test tests[] = {
     T(digest_failures),
     T(fail_challenge),
     T(multi_handler),
+    T(domains),
+    T(defaults),
     T(NULL)
 };
